"""
BGP test functions
"""

from typing import Any, Dict, Optional

from anta.decorators import check_bgp_family_enable
from anta.models import AntaCommand, AntaTemplate, AntaTest


def _check_bgp_vrfs(bgp_vrfs: Dict[str, Any]) -> Dict[str, Any]:
    """
    TODO
    """
    state_issue: Dict[str, Any] = {}
    for vrf in bgp_vrfs:
        for peer in bgp_vrfs[vrf]["peers"]:
            if (
                (bgp_vrfs[vrf]["peers"][peer]["peerState"] != "Established")
                or (bgp_vrfs[vrf]["peers"][peer]["inMsgQueue"] != 0)
                or (bgp_vrfs[vrf]["peers"][peer]["outMsgQueue"] != 0)
            ):
                vrf_dict = state_issue.setdefault(vrf, {})
                vrf_dict.update(
                    {
                        peer: {
                            "peerState": bgp_vrfs[vrf]["peers"][peer]["peerState"],
                            "inMsgQueue": bgp_vrfs[vrf]["peers"][peer]["inMsgQueue"],
                            "outMsgQueue": bgp_vrfs[vrf]["peers"][peer]["outMsgQueue"],
                        }
                    }
                )

    return state_issue


class VerifyBGPIPv4UnicastState(AntaTest):
    """
    Verifies all IPv4 unicast BGP sessions are established (for all VRF)
    and all BGP messages queues for these sessions are empty (for all VRF).

    * self.result = "skipped" if no BGP vrf are returned by the device
    * self.result = "success" if all IPv4 unicast BGP sessions are established (for all VRF)
                         and all BGP messages queues for these sessions are empty (for all VRF).
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPIPv4UnicastState"
    description = "Verifies all IPv4 unicast BGP sessions are established (for all VRF) and all BGP messages queues for these sessions are empty (for all VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp ipv4 unicast summary vrf all")]

    @check_bgp_family_enable("ipv4")
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyBGPIPv4UnicastState validation"""

        command_output = self.instance_commands[0].json_output
        state_issue = _check_bgp_vrfs(command_output["vrfs"])

        if not state_issue:
            self.result.is_success()
        else:
            self.result.is_failure(f"Some IPv4 Unicast BGP Peer are not up: {state_issue}")


class VerifyBGPIPv4UnicastCount(AntaTest):
    """
    Verifies all IPv4 unicast BGP sessions are established
    and all BGP messages queues for these sessions are empty
    and the actual number of BGP IPv4 unicast neighbors is the one we expect.

    * self.result = "skipped" if the `number` or `vrf` parameter is missing
    * self.result = "success" if all IPv4 unicast BGP sessions are established
                         and if all BGP messages queues for these sessions are empty
                         and if the actual number of BGP IPv4 unicast neighbors is equal to `number.
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPIPv4UnicastCount"
    description = (
        "Verifies all IPv4 unicast BGP sessions are established and all their BGP messages queues are empty and "
        " the actual number of BGP IPv4 unicast neighbors is the one we expect."
    )
    categories = ["routing", "bgp"]
    template = AntaTemplate(template="show bgp ipv4 unicast summary vrf {vrf}")

    @check_bgp_family_enable("ipv4")
    @AntaTest.anta_test
    def test(self, number: Optional[int] = None) -> None:
        """
        Run VerifyBGPIPv4UnicastCount validation

        Args:
            number: The expected number of BGP IPv4 unicast neighbors.
            vrf: VRF to verify (template parameter)
        """

        if not number:
            self.result.is_skipped("VerifyBGPIPv4UnicastCount could not run because number was not supplied")
            return

        self.result.is_success()

        for command in self.instance_commands:
            if command.params and "vrf" in command.params:
                vrf = command.params["vrf"]

            peers = command.json_output["vrfs"][vrf]["peers"]
            state_issue = _check_bgp_vrfs(command.json_output["vrfs"])

            if len(peers) != number:
                self.result.is_failure(f"Expecting {number} BGP peer in vrf {vrf} and got {len(peers)}")
            if state_issue:
                self.result.is_failure(f"The following IPv4 peers are not established: {state_issue}")


class VerifyBGPIPv6UnicastState(AntaTest):
    """
    Verifies all IPv6 unicast BGP sessions are established (for all VRF)
    and all BGP messages queues for these sessions are empty (for all VRF).

    * self.result = "skipped" if no BGP vrf are returned by the device
    * self.result = "success" if all IPv6 unicast BGP sessions are established (for all VRF)
                         and all BGP messages queues for these sessions are empty (for all VRF).
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPIPv6UnicastState"
    description = "Verifies all IPv6 unicast BGP sessions are established (for all VRF) and all BGP messages queues for these sessions are empty (for all VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp ipv6 unicast summary vrf all")]

    @check_bgp_family_enable("ipv6")
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyBGPIPv6UnicastState validation"""

        command_output = self.instance_commands[0].json_output

        state_issue = _check_bgp_vrfs(command_output["vrfs"])

        if not state_issue:
            self.result.is_success()
        else:
            self.result.is_failure(f"Some IPv4 Unicast BGP Peer are not up: {state_issue}")


class VerifyBGPEVPNState(AntaTest):
    """
    Verifies all EVPN BGP sessions are established (default VRF).

    * self.result = "skipped" if no BGP EVPN peers are returned by the device
    * self.result = "success" if all EVPN BGP sessions are established.
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPEVPNState"
    description = "Verifies all EVPN BGP sessions are established (default VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp evpn summary")]

    @check_bgp_family_enable("evpn")
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyBGPEVPNState validation"""

        command_output = self.instance_commands[0].json_output

        bgp_vrfs = command_output["vrfs"]

        peers = bgp_vrfs["default"]["peers"]
        non_established_peers = [peer for peer, peer_dict in peers.items() if peer_dict["peerState"] != "Established"]

        if not non_established_peers:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following EVPN peers are not established: {non_established_peers}")


class VerifyBGPEVPNCount(AntaTest):
    """
    Verifies all EVPN BGP sessions are established (default VRF)
    and the actual number of BGP EVPN neighbors is the one we expect (default VRF).

    * self.result = "skipped" if the `number` parameter is missing
    * self.result = "success" if all EVPN BGP sessions are Established and if the actual
                         number of BGP EVPN neighbors is the one we expect.
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPEVPNCount"
    description = "Verifies all EVPN BGP sessions are established (default VRF) and the actual number of BGP EVPN neighbors is the one we expect (default VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp evpn summary")]

    @check_bgp_family_enable("evpn")
    @AntaTest.anta_test
    def test(self, number: Optional[int] = None) -> None:
        """
        Run VerifyBGPEVPNCount validation

        Args:
            number: The expected number of BGP EVPN neighbors in the default VRF.
        """
        if not number:
            self.result.is_skipped("VerifyBGPEVPNCount could not run because number was not supplied.")
            return

        command_output = self.instance_commands[0].json_output

        peers = command_output["vrfs"]["default"]["peers"]
        non_established_peers = [peer for peer, peer_dict in peers.items() if peer_dict["peerState"] != "Established"]

        if not non_established_peers and len(peers) == number:
            self.result.is_success()
        else:
            self.result.is_failure()
            if len(peers) != number:
                self.result.is_failure(f"Expecting {number} BGP EVPN peers and got {len(peers)}")
            if non_established_peers:
                self.result.is_failure(f"The following EVPN peers are not established: {non_established_peers}")


class VerifyBGPRTCState(AntaTest):
    """
    Verifies all RTC BGP sessions are established (default VRF).

    * self.result = "skipped" if no BGP RTC peers are returned by the device
    * self.result = "success" if all RTC BGP sessions are established.
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPRTCState"
    description = "Verifies all RTC BGP sessions are established (default VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp rt-membership summary")]

    @check_bgp_family_enable("rtc")
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyBGPRTCState validation"""

        command_output = self.instance_commands[0].json_output

        bgp_vrfs = command_output["vrfs"]

        peers = bgp_vrfs["default"]["peers"]
        non_established_peers = [peer for peer, peer_dict in peers.items() if peer_dict["peerState"] != "Established"]

        if not non_established_peers:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following RTC peers are not established: {non_established_peers}")


class VerifyBGPRTCCount(AntaTest):
    """
    Verifies all RTC BGP sessions are established (default VRF)
    and the actual number of BGP RTC neighbors is the one we expect (default VRF).

    * self.result = "skipped" if the `number` parameter is missing
    * self.result = "success" if all RTC BGP sessions are Established and if the actual
                         number of BGP RTC neighbors is the one we expect.
    * self.result = "failure" otherwise.
    """

    name = "VerifyBGPRTCCount"
    description = "Verifies all RTC BGP sessions are established (default VRF) and the actual number of BGP RTC neighbors is the one we expect (default VRF)."
    categories = ["routing", "bgp"]
    commands = [AntaCommand(command="show bgp rt-membership summary")]

    @check_bgp_family_enable("rtc")
    @AntaTest.anta_test
    def test(self, number: Optional[int] = None) -> None:
        """
        Run VerifyBGPRTCCount validation

        Args:
            number: The expected number of BGP RTC neighbors (default VRF).
        """
        if not number:
            self.result.is_skipped("VerifyBGPRTCCount could not run because number was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        peers = command_output["vrfs"]["default"]["peers"]
        non_established_peers = [peer for peer, peer_dict in peers.items() if peer_dict["peerState"] != "Established"]

        if not non_established_peers and len(peers) == number:
            self.result.is_success()
        else:
            self.result.is_failure()
            if len(peers) != number:
                self.result.is_failure(f"Expecting {number} BGP RTC peers and got {len(peers)}")
            if non_established_peers:
                self.result.is_failure(f"The following RTC peers are not established: {non_established_peers}")



"""Models related to anta.result_manager module."""

from typing import Iterator, List, Optional

from pydantic import BaseModel, RootModel, field_validator

RESULT_OPTIONS = ["unset", "success", "failure", "error", "skipped"]


class TestResult(BaseModel):
    """
    Describe the result of a test from a single device.

    Attributes:
        name (str): Device name where the test has run.
        test (str): Test name runs on the device.
        categories (List[str]): List of categories the TestResult belongs to, by default the AntaTest categories.
        description (str): TestResult description, by default the AntaTest description.
        results (str): Result of the test. Can be one of ["unset", "success", "failure", "error", "skipped"].
        message (str, optional): Message to report after the test if any.
        custom_field (str, optional): Custom field to store a string for flexibility in integrating with ANTA
    """

    name: str
    test: str
    categories: List[str]
    description: str
    result: str = "unset"
    messages: List[str] = []
    custom_field: Optional[str] = None

    @classmethod
    @field_validator("result")
    def name_must_be_in(cls, v: str) -> str:
        """
        Status validator

        Validate status is a supported one

        Args:
            v (str): User defined status

        Raises:
            ValueError: If status is unsupported

        Returns:
            str: status value
        """
        if v not in RESULT_OPTIONS:
            raise ValueError(f"must be one of {RESULT_OPTIONS}")
        return v

    def is_success(self, message: str = "") -> bool:
        """
        Helper to set status to success

        Args:
            message (str): Optional message related to the test

        Returns:
            bool: Always true
        """
        return self._set_status("success", message)

    def is_failure(self, message: str = "") -> bool:
        """
        Helper to set status to failure

        Args:
            message (str): Optional message related to the test

        Returns:
            bool: Always true
        """
        return self._set_status("failure", message)

    def is_skipped(self, message: str = "") -> bool:
        """
        Helper to set status to skipped

        Args:
            message (str): Optional message related to the test

        Returns:
            bool: Always true
        """
        return self._set_status("skipped", message)

    def is_error(self, message: str = "") -> bool:
        """
        Helper to set status to error

        Args:
            message (str): Optional message related to the test

        Returns:
            bool: Always true
        """
        return self._set_status("error", message)

    def _set_status(self, status: str, message: str = "") -> bool:
        """
        Set status and insert optional message

        Args:
            status (str): status of the test
            message (str): optional message

        Returns:
            bool: Always true
        """
        self.result = status
        if message != "":
            self.messages.append(message)
        return True

    def __str__(self) -> str:
        """
        Returns a human readable string of this TestResult
        """
        return f"Test {self.test} on device {self.name} has result {self.result}"


class ListResult(RootModel[List[TestResult]]):
    """
    List result for all tests on all devices.

    Attributes:
        __root__ (List[TestResult]): A list of TestResult objects.
    """

    root: List[TestResult] = []

    def extend(self, values: List[TestResult]) -> None:
        """Add support for extend method."""
        self.root.extend(values)

    def append(self, value: TestResult) -> None:
        """Add support for append method."""
        self.root.append(value)

    def __iter__(self) -> Iterator[TestResult]:  # type: ignore
        """Use custom iter method."""
        # TODO - mypy is not happy because we overwrite BaseModel.__iter__
        # return type and are breaking Liskov Substitution Principle.
        return iter(self.root)

    def __getitem__(self, item: int) -> TestResult:
        """Use custom getitem method."""
        return self.root[item]

    def __len__(self) -> int:
        """Support for length of __root__"""
        return len(self.root)



"""decorators for tests."""
from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

from anta.models import AntaCommand, AntaTest
from anta.tools.misc import exc_to_str

if TYPE_CHECKING:
    from anta.result_manager.models import TestResult

# TODO - should probably use mypy Awaitable in some places rather than this everywhere - @gmuloc
F = TypeVar("F", bound=Callable[..., Any])


def skip_on_platforms(platforms: list[str]) -> Callable[[F], F]:
    """Return a decorator factory to skip a test on a list of platforms.

    Args:
    ----
      platforms (List[str]): the list of platforms on which the decorated test should be skipped.

    """

    def decorator(function: F) -> F:
        """Implement the decorator to skip a test ona list of platform.

        Args:
        ----
            function (F): the test to be decorated.
        """

        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> TestResult:
            """Implement the wrapper for func."""
            anta_test = args[0]

            if anta_test.result.result != "unset":
                AntaTest.update_progress()
                return anta_test.result

            if anta_test.device.hw_model in platforms:
                anta_test.result.is_skipped(f"{anta_test.__class__.__name__} test is not supported on {anta_test.device.hw_model}.")
                AntaTest.update_progress()
                return anta_test.result

            return await function(*args, **kwargs)

        return cast(F, wrapper)

    return decorator


def check_bgp_family_enable(family: str) -> Callable[[F], F]:
    """Return a decorator factory to skip a test if BGP is enabled.

    Args:
    ----
        family (str): BGP family to check. Can be ipv4 / ipv6 / evpn / rtc

    """

    def decorator(function: F) -> F:
        """Implement the decorator to skip a test if an address family is not configured.

        Args:
        ----
            function (F): the test to be decorated.
        """

        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> TestResult:
            """Implement the wrapper for func."""
            anta_test = args[0]

            if anta_test.result.result != "unset":
                AntaTest.update_progress()
                return anta_test.result

            if family == "ipv4":
                command = AntaCommand(command="show bgp ipv4 unicast summary vrf all")
            elif family == "ipv6":
                command = AntaCommand(command="show bgp ipv6 unicast summary vrf all")
            elif family == "evpn":
                command = AntaCommand(command="show bgp evpn summary")
            elif family == "rtc":
                command = AntaCommand(command="show bgp rt-membership summary")
            else:
                anta_test.result.is_error(f"Wrong address family for bgp decorator: {family}")
                return anta_test.result

            await anta_test.device.collect(command=command)

            if not command.collected and command.failed is not None:
                anta_test.result.is_error(f"{command.command}: {exc_to_str(command.failed)}")
                return anta_test.result
            if "vrfs" not in command.json_output:
                anta_test.result.is_skipped(f"no BGP configuration for {family} on this device")
                AntaTest.update_progress()
                return anta_test.result
            if len(bgp_vrfs := command.json_output["vrfs"]) == 0 or len(bgp_vrfs["default"]["peers"]) == 0:
                # No VRF
                anta_test.result.is_skipped(f"no {family} peer on this device")
                AntaTest.update_progress()
                return anta_test.result

            return await function(*args, **kwargs)

        return cast(F, wrapper)

    return decorator



"""
Loader that parses a YAML test catalog and imports corresponding Python functions
"""
import importlib
import logging
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

from rich.logging import RichHandler

from anta import __DEBUG__
from anta.result_manager.models import TestResult

logger = logging.getLogger(__name__)


def setup_logging(level: str = logging.getLevelName(logging.INFO)) -> None:
    """
    Configure logging

    Args:
        level (str, optional): level name to configure.
    """
    root = logging.getLogger()
    loglevel = getattr(logging, level.upper())
    handler = RichHandler(markup=True, rich_tracebacks=True, tracebacks_show_locals=True)
    if __DEBUG__:
        fmt_string = r"[grey58]\[%(name)s][/grey58] %(message)s"
    else:
        fmt_string = "%(message)s"
    formatter = logging.Formatter(fmt=fmt_string, datefmt="[%X]")
    handler.setFormatter(formatter)
    root.addHandler(handler)
    root.setLevel(loglevel)

    if __DEBUG__:
        if loglevel != logging.DEBUG:
            root.setLevel(logging.DEBUG)
            logger.debug(f"Override current logging level {level.upper()} with DEBUG")
        logger.debug("ANTA Debug Mode enabled")
    else:
        if loglevel == logging.INFO:
            # asyncssh is really chatty
            logging.getLogger("asyncssh").setLevel(logging.WARNING)
            # httpx as well
            logging.getLogger("httpx").setLevel(logging.WARNING)


def parse_catalog(test_catalog: Dict[Any, Any], package: Optional[str] = None) -> List[Tuple[Callable[..., TestResult], Dict[Any, Any]]]:
    """
    Function to parse the catalog and return a list of tests

    Args:
        test_catalog (Dict[Any, Any]): List of tests defined in catalog YAML file

    Returns:
        List[Tuple[Callable[..., TestResult], Dict[Any, Any]]]: List of python function tests to run.
    """
    tests: List[Tuple[Callable[..., TestResult], Dict[Any, Any]]] = []
    if not test_catalog:
        return tests
    for key, value in test_catalog.items():
        # Required to manage iteration within a tests module
        if package is not None:
            key = ".".join([package, key])
        try:
            module = importlib.import_module(f"{key}")
        except ModuleNotFoundError:
            logger.error(f"No test module named '{key}'")
            sys.exit(1)
        if isinstance(value, list):
            # This is a list of tests
            for test in value:
                for func_name, args in test.items():
                    try:
                        func = getattr(module, func_name)
                    except AttributeError:
                        logger.error(f"Wrong test function name '{func_name}' in '{module.__name__}'")
                        sys.exit(1)
                    if not callable(func):
                        logger.error(f"'{func.__module__}.{func.__name__}' is not a function")
                        sys.exit(1)
                    tests.append((func, args if args is not None else {}))
        if isinstance(value, dict):
            # This is an inner Python module
            tests.extend(parse_catalog(value, package=module.__name__))
    return tests
